(* Bantam -- a lightweight file manager for X11
   Copyright  2005 by Matthew C. Gushee

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)

open Common
open GdkKeysyms


let prefs =
  let user_prefs_file = Filename.concat BConfig.config_dir "prefs" in
  let pobj = Prefs.prefs [user_prefs_file] "" in
  ( try
      pobj#load ()
    with _ ->
      ( prerr_endline ("WARNING: Your preferences file, " ^ user_prefs_file ^
                       ", could not be loaded. This means that Bantam may " ^
                       "not be correctly installed. Continuing with default " ^
                       "configuration.");
        flush stderr ) );
  pobj

class type app_t = object
  method w : GWindow.window
  method p : GPack.box
  method viewpanel : viewpanel_t
  method workbench : workbench_t
  method tooltips : GData.tooltips
  method defaults : defaults_t
  method guifont : Pango.font_description
  method textfont : Pango.font_description
  method tipfont : Pango.font_description
  method command : string -> ((dirpanel_t -> unit) * string)
  method keymap : ((Gdk.keysym * bool) * string) list
  method viewer : string -> handler
  method editor : string -> handler
  method warn : string -> unit
  method get_model : string -> dirmodel_t
  method update_views : string -> unit
  method next_panel : dirpanel_t -> dirpanel_t option
  method prev_panel : dirpanel_t -> dirpanel_t option
  (* method move_to : dirpanel_t -> unit *)
  method set_active : dirpanel_t -> GdkEvent.Focus.t -> bool
  method unset_active : GdkEvent.Focus.t -> bool
  method set_last : dirpanel_t -> GdkEvent.Focus.t -> bool
  method int_close_viewpanel : unit -> unit
  method int_add_panel : ?dir:string -> ?go:bool -> unit -> unit
  method int_view_file : string -> unit
  method int_view_text : string -> string -> unit
  method show_help : unit -> unit
  method init : ?first:bool -> ?daemon:bool -> unit -> unit
  method show : unit -> unit
  method hide : unit -> unit
  method quit : unit -> unit
  method first_time_greeting : unit -> unit
end
and dirpanel_t = object
  method id : int
  method hull : GPack.box
  method size : int * int
  method directory : string
  method set_directory : string -> unit
  method view : GTree.view
  method vcol : GTree.view_column
  method set_vcol : GTree.view_column -> unit
  method model : dirmodel_t
  method set_model : dirmodel_t -> unit
  method unset_model : unit -> unit
  method update : unit -> unit
  (*
  method get_input : ?start:Tk.entry_index -> ?stop:Tk.entry_index ->
      ?default:string -> string -> string option
  *)
  (* Should I multiplex the parameters? Not needed just now, but ... *)
  method do_with_param : ?default:string -> ?start:int -> ?stop:int ->
    string -> (string -> unit) -> unit
  method save_selected : unit -> unit
  method set_selected : ?force:Gtk.tree_path -> unit -> unit
  method set_focus : ?force:Gtk.tree_path -> unit -> bool
  method selected_file : selection
  method copy_to : ?name:string -> string -> unit
  method move_to : ?name:string -> string -> bool
  method destroy : unit -> unit
end
and viewpanel_t = object
  method hull : GPack.box
  method size : int * int
  method set_focus : ?force:GText.iter -> unit -> bool
  method clear : unit -> unit
  method load_file : string -> unit
  method load_text : string -> string -> unit
  (* method tag_configure : string -> GText.tag_property list -> unit *)
  method load_tagtext :
    ?is_file:bool -> string -> (string * GText.tag_property list) list ->
    (string * string) list -> unit
  method first_time_greeting : unit -> unit
end
and workbench_t = object
  method hull : GPack.box
  method set_focus : unit -> unit
end
and microdialog_t = object
  method interact : ?start:int -> ?stop:int ->
    string -> string -> (unit -> unit) -> unit
end
and dirmodel_t = object
  method dir : string
  method mtime : float
  (*
  method vcol : int -> GTree.view_column
  method remove_vcol : int -> unit
  *)
  (*
  method vcol : GTree.view_column
  *)
  method toggle_hidden : unit -> unit
  method update : unit -> unit
  (*
  method attach_to_view : int -> GTree.view -> unit
  method detach_from_view : int -> GTree.view -> unit
  *)
  method attach_to_panel : dirpanel_t -> unit
  method detach_from_panel : dirpanel_t -> unit
  method get_text : Gtk.tree_path list -> string
end



class text_variable = object(self)
  val mainvar = new GUtil.variable ""
  val oldvar = new GUtil.variable None
  val mutable sub_callbacks:(string -> unit) list = []
  val mutable initialized = false
  (*
  val mutable null_ = true
  *)
  
  initializer
    ignore (mainvar#connect#changed ~callback:self#do_sub_callbacks)

  method private do_sub_callbacks newval =
    List.iter (fun cb -> cb newval) sub_callbacks

  method add_callback cb = sub_callbacks <- cb :: sub_callbacks
  (*
  method is_null = null_
  *)
  method get =
    if initialized then mainvar#get
    else failwith "Textvariable must be initialized before first use."
  method get_prev =
    if initialized then oldvar#get
    else None
  method set newval =
    if initialized then
      ( oldvar#set (Some mainvar#get); mainvar#set newval )
    else
      ( oldvar#set None; mainvar#set newval; initialized <- true )
    
  (*
  method set newval =
    var#set newval;
    null_ <- false
  method unset () =
    var#set "";
    null_ <- true
  *)
end


class microdialog app (master:GPack.box) =

    let hull = GPack.hbox () in
    let label = GMisc.label ~text:"[prompt]" ~packing:hull#add ()
    and entry = GEdit.entry ~packing:hull#add () in

  object(self)
(*
  method private set_width () =
    let ws = tkEval
      [|cCAMLtoTKwidget self#inputw; TkToken "cget"; TkToken "-width"|] in
    (* prerr_endline ("WIDTH: " ^ ws); *)
    let w = int_of_string ws in
    width <- w
*)

    val mutable callback_ = fun (s:string) -> ()
  
    initializer
      ignore (entry#event#connect#key_release ~callback:self#on_key_release)
  
    method private on_key_release ev =
      let sym = GdkEvent.Key.keyval ev in
      if sym = GdkKeysyms._Return then (self#ok (); true)
      else if sym = GdkKeysyms._Escape then (self#cancel (); true)
      else false
    method private show ?start ?stop () =
      let start' =
        match start with
        | Some n -> n
        | None -> 0
      and stop' =
        match stop with
        | Some n -> min n (entry#text_length)
        | None -> entry#text_length in
      master#pack hull#coerce;
      entry#select_region ~start:start' ~stop:stop';
      entry#misc#grab_focus ()
    method private hide () : unit =
      master#remove hull#coerce 
    method private ok () = 
      self#hide ();
      callback_ entry#text
    method private cancel () = 
      self#hide ()
    method interact ?start ?stop prompt initial callback : unit = 
      label#set_text (prompt ^ ": ");
      entry#set_text initial;
      callback_ <- callback;
      self#show ?start ?stop ()
end


(*
module CT = Weak.Make (
  struct
    type t = int * GTree.view_column
    let equal (id1, _) (id2, _) = (id1 = id2)
    let hash (id, _) = Hashtbl.hash id
  end
)
*)
(*
module CTBase = Weak.Make (
  struct
    type t = int * GTree.view_column
    let equal (id1, _) (id2, _) = (id1 = id2)
    let hash (id, _) = Hashtbl.hash id
  end
)
module CT = struct
  include CTBase
  let dump ct =
    print_endline "Column table:\n=============\n"; 
    (* Printf.printf "Size: %d\n\n" (count ct); *)
    print_endline "Columns:\n--------";
    (* iter (fun (id, _) -> print_int id; print_newline ()) ct; *)
    print_newline ();
    flush stdout
end
*)
(*
module CT : sig
  type ('a, 'b) t
  val create : int -> ('a, 'b) t
  val add : ('a, 'b) t -> 'a * 'b -> unit
  val find : ('a, 'b) t -> 'a * 'b -> 'a * 'b 
  val remove : ('a, 'b) t -> 'a * 'b -> 
end = struct
  type ('a, 'b) t = ('a, 'b) Hashtbl.t 
  let create = Hashtbl.create
  let add ct (k, v) = Hashtbl.add ct k v
  let find ct (key, _) = (key, Hashtbl.find ct key)
end
*)

class dir_model ?(show_hidden=false) dir =

    let cols = new GTree.column_list in
    let col_f = cols#add Gobject.Data.string in 
    let store = GTree.list_store cols
    and ren = GTree.cell_renderer_text []
    and dummy_vcol = GTree.view_column () in

  object(self)
    val mutable show_hidden_ = show_hidden
    val mutable mtime_ = 0.
    (* val vcols = CT.create 11 *)

    initializer
      self#update ()

    method coerce = (self :> dirmodel_t)
    method dir = dir
    method mtime = mtime_
        
    method toggle_hidden () =
      show_hidden_ <- not show_hidden_;
      self#update ()
    method update () = 
      let fill fname =
        let iter = store#append () in
        store#set ~row:iter ~column:col_f
          (Glib.Convert.locale_to_utf8 fname)
      and files, time = Util.list_files ~show_hidden:show_hidden_ dir in
      store#clear ();
      List.iter fill files;
      mtime_ <- time
    method attach_to_panel p =
      let v = p#view in
      v#set_model (Some store#coerce);
      let vc = GTree.view_column ~renderer:(ren, ["text", col_f]) () in
      p#set_vcol vc;
      ignore (v#append_column vc)
    method detach_from_panel p =
      let v = p#view
      and vc = p#vcol in
      ignore (v#remove_column vc);
      vc#destroy ()
    method get_text paths =
      let do_get prev path =
        let row = store#get_iter path in
        store#get ~row ~column:col_f in
      if List.length paths > 1 then
        Util.warn "More than one row is selected. That should not happen.";
      List.fold_left do_get "" paths
      
end

class dirpanel (app:app_t) id =
    let dirvar = new text_variable in
    let hull = GPack.vbox ~packing:app#p#pack () in
    let label = GBin.event_box ~height:16 ~packing:hull#pack () in
    let lbl_text = GMisc.label ~xpad:1 ~ypad:1 ~packing:label#add ()
    and lw = GBin.scrolled_window ~width:prefs#panel_width
      ~shadow_type:`IN
      ~height:prefs#panel_height ~hpolicy:`NEVER
      ~vpolicy:`AUTOMATIC ~packing:(hull#pack ~expand:true) () in
    let listbox = GTree.view ~width:(prefs#panel_width-lw#border_width)
      ~height:(prefs#panel_height-lw#border_width)
      (* ~packing:lw#add_with_viewport *)
      ~packing:lw#add
      ~headers_visible:false ~enable_search:false () in
    let microdia = new microdialog app hull in

  object(self)
  
    val id_ = id
    val mutable directory_ = ""
    val selected_:(string, Gtk.tree_path option) Hashtbl.t = Hashtbl.create 23
      (* 37, 47? *)
    val mutable too_wide = false
    val path_display_width = 24
    val mutable show_hidden = app#defaults.show_hidden
    (* val mutable vcol_:GTree.view_column option = None *)
    val mutable model_:dirmodel_t option = None
    val mutable width_ = prefs#panel_width
    val mutable height_ = prefs#panel_height
    val mutable vcol = GTree.view_column ()
    (* This a kludge to prevent a column from being set as last_visited
       when it is destroyed *)
    val mutable destroyed = false
  
    initializer
      listbox#selection#set_mode `BROWSE;
      dirvar#add_callback self#on_chdir;
      self#set_directory (Sys.getcwd ());
      self#set_bindings ()
  
    method private set_bindings () =
      (* Can't bind to key_release because some (unwanted) default
         actions are bound to key_press *)
      ignore (listbox#event#connect#key_press ~callback:self#key_handler);
      ignore (listbox#event#connect#button_release ~callback:self#take_focus);
      ignore (listbox#event#connect#focus_in ~callback:(self#set_active app));
      ignore (listbox#event#connect#focus_out ~callback:self#set_last)

    (* DEBUGGING ONLY *)
    method private on_destroy ev =
      prerr_endline "Ouch! I'm being destroyed!"; flush stderr;
      true
  
    (* Widget access *)
    method id = id_
    method hull = hull
    method model =
      match model_ with
      | Some m -> m
      | None -> raise (IllegalState "directory model not initialized")
    method view = listbox
    method private lb = listbox
    method private microdia = microdia
    method vcol = vcol
    method set_vcol vc = vcol <- vc
    method private coerce = (self :> dirpanel_t)
    method size = width_, height_
  
    (* Utility methods *)
    method directory = dirvar#get
    method private prev_dir = dirvar#get_prev
    method selected_file =
      let dir = self#directory in
      let paths = listbox#selection#get_selected_rows in
      try
        let raw = self#model#get_text paths in
        if raw = ".. " then ParentDir
        else
          let base = String.sub raw 0 (String.length raw - 1) in
          File (Filename.concat dir base)
      with _ -> NoSelection
    method private listsel dir =
      try Hashtbl.find selected_ dir
      with Not_found -> None
    method private listsel_set dir paths =
      let path =
        match paths with
        | [] -> None
        | [p] -> Some p
        | p::_ -> (app#warn "Multiple rows selected."; Some p) in
      Hashtbl.replace selected_ dir path

    method private set_last (ev:GdkEvent.Focus.t) =
      if destroyed then true
      else app#set_last self#coerce ev
    (*
    method get_input ?start ?stop ?(default="") prompt =
      self#microdia#get_input ?start ?stop default prompt
    *)
    method do_with_param ?(default="") ?(start=0) ?stop prompt callback =
      self#microdia#interact ~start:start ?stop
        prompt default callback

    method set_directory dir =
      dirvar#set dir
    method set_model dm =
      model_ <- Some dm
    method unset_model () =
      model_ <- None
  
    method copy_to ?name file : unit =
      let dest =
        match name with
        | Some nm -> nm
        | None -> self#directory in
      let cmd =
        match Util.get_filetype file with
        | Directory -> Printf.sprintf "cp -R %s %s" file dest
        | _ -> Printf.sprintf "cp %s %s" file dest in
      let code = Sys.command cmd in
      (* if code = 0 then self#update () *)
      if code = 0 then app#update_views self#directory
      else app#warn "Failed to copy file."
    method move_to ?name file =
      let dest =
        match name with
        | Some nm -> nm
        | None -> self#directory in
      let cmd = Printf.sprintf "mv %s %s" file dest in
      let code = Sys.command cmd in
      (* if code = 0 then (self#update (); true) *)
      if code = 0 then (app#update_views self#directory; true)
      else (app#warn "Failed to move file."; false)
  
    (* Event handlers *)
    method private on_chdir dir =
      ( match self#prev_dir with
        | None -> ()
        | Some d ->
          let old_dm = app#get_model d in
          old_dm#detach_from_panel self#coerce );
      let dm = app#get_model dir in
      self#set_model dm;
      dm#attach_to_panel self#coerce;
      app#tooltips#set_tip ~text:dir label#coerce;
      let labeltxt = Util.chop_dir dir path_display_width in
      lbl_text#set_text labeltxt
    method private key_handler ev =
      let ks = GdkEvent.Key.keyval ev
      and shift = List.mem `SHIFT (GdkEvent.Key.state ev) in
      if ks = _period then (self#toggle_hidden self#lb; true)
      (*
      else if ks = _Prior then (self#sync_selection self#lb; true)
      else if ks = _Next then (self#sync_selection self#lb; true)
      *)
      else
        ( try 
            ( let cmdname = BConfig.get_key_cmd (ks, shift) app#keymap in
              let cmd, _ = app#command cmdname in
              cmd (self :> dirpanel_t);
              true )
          with Not_found -> false )
  
    method private toggle_hidden info =
      self#model#toggle_hidden ()
  
    method update () =
      self#save_selected ();
      self#model#update ()
  
    method save_selected () =
      try self#listsel_set self#directory listbox#selection#get_selected_rows
      with _ -> ()
  
    method set_selected ?force () =
      let pth =
        match force with
        | Some path ->
          self#listsel_set self#directory [path]; path
        | None ->
          ( let sel = self#listsel self#directory in
            match sel with
            | Some p -> p
            | None -> GTree.Path.create [0] ) in
      (* Don't see how to do this w/ GTK *)
      (* Listbox.see ~index:sel' self#lb *)
      listbox#selection#select_path pth;
      listbox#scroll_to_cell ~align:(0.5,0.) pth vcol
  
    method private take_focus ev =
      let x = Util.round_int (GdkEvent.Button.x ev)
      and y = Util.round_int (GdkEvent.Button.y ev) in
      let pathinf = listbox#get_path_at_pos x y in
      match pathinf with
      | None -> true
      | Some (path, _, _, _) -> self#set_focus ~force:path ()
  
    method set_focus ?force () =
      listbox#misc#grab_focus ();
      self#set_selected ?force ();
      true
  
    method private set_active app ev =
      ignore (app#set_active (self :> dirpanel_t) ev);
      self#set_selected ();
      true
  
    (* FIXME: not sure how to do this, or if it's really needed. *)
    method private sync_selection einf =
      ()
      (*
      Listbox.selection_clear ~first:(`Num 0) ~last:`End self#lb;
      Listbox.selection_set ~first:`Active ~last:`Active self#lb;
      Listbox.see ~index:`Active self#lb
      *)
    method destroy () =
      destroyed <- true;
      let dm = app#get_model self#directory in
      dm#detach_from_panel self#coerce;
      hull#destroy ()
end

class viewpanel app =
    let titlevar = new text_variable in
    let hull = GPack.vbox () in
    let label = GBin.event_box ~height:16 ~packing:hull#pack () in
    let lbl_text = GMisc.label ~xpad:1 ~ypad:1 ~packing:label#add ()
    and sw = GBin.scrolled_window ~hpolicy:`AUTOMATIC ~vpolicy:`AUTOMATIC
      ~width:prefs#viewer_width ~height:prefs#panel_height
      ~packing:(hull#pack ~expand:true) () in
    let textwin = GText.view ~editable:false ~wrap_mode:`WORD
      ~packing:sw#add_with_viewport () in
  object(self)
    val mutable have_file = false
    val mutable width_ = prefs#viewer_width
    val mutable height_ = prefs#panel_height
  
    initializer
      titlevar#add_callback self#on_chtitle;
      lbl_text#misc#modify_font app#guifont; 
      textwin#misc#modify_font app#textfont;
      self#set_bindings ()
  
    method private set_bindings () =
      ignore (textwin#event#connect#key_release ~callback:self#key_handler);
      ignore (textwin#event#connect#button_release ~callback:self#take_focus);
      ignore (textwin#event#connect#focus_in ~callback:app#unset_active)

    method private key_handler ev =
      let ks = GdkEvent.Key.keyval ev in
      if ks = _a || ks = _A || ks = _q || ks = _Q then
        ( app#int_close_viewpanel (); true )
      else if ks = _e then ( self#edit_file (); true )
      else false
  
    method private tw = textwin
  
    method hull = hull

    method size = width_, height_
  
    method set_focus ?force () =
      (* XXX Does this work given that it's not set to editable? XXX *)
      textwin#misc#grab_focus ();
      match force with
      | Some iter -> textwin#scroll_to_iter iter
      | None -> true
    method private take_focus ev =
      let x = Util.round_int (GdkEvent.Button.x ev)
      and y = Util.round_int (GdkEvent.Button.y ev) in
      let iter = textwin#get_iter_at_location ~x:x ~y:y in 
      self#set_focus ~force:iter ()

    (* XXX Maybe unnecessary? XXX *) 
    method clear () =
      (* Text.configure ~state:`Normal self#tw; *)
      let start = textwin#buffer#get_iter `START
      and stop = textwin#buffer#get_iter `END in
      textwin#buffer#delete ~start:start ~stop:stop;
      titlevar#set "";
      have_file <- false
    method private on_chtitle title =
      lbl_text#set_text title
    method private load title content =
      textwin#buffer#set_text content;
      let i = textwin#buffer#get_iter `START in
      ignore (textwin#scroll_to_iter i);
      titlevar#set title
    method load_file fname =
      self#load fname (Util.read_file fname); have_file <- true
    method load_text title content =
      self#load title content;
      have_file <- false
    method load_tagtext ?(is_file=false) title tag_props texts =
      titlevar#set title;
      let tagmap =
        List.map
          ( fun (name, props) -> name, textwin#buffer#create_tag props )
          tag_props in
      List.iter
        ( fun (tagname, txt) ->
          try
            let tag = List.assoc tagname tagmap in
            textwin#buffer#insert ~tags:[tag] txt
          with Not_found ->
            ( app#warn ("Tag '" ^ tagname ^ "' not found in viewpanel.");
              textwin#buffer#insert txt ) )
        texts;
      have_file <- is_file
    (* method private edit_file einf = *)
    method private edit_file () =
      if have_file then
        ( let fname = titlevar#get in
          let ed = app#editor fname in
          match ed with
          | ExtHandler edname ->
            ( let cmd = Printf.sprintf "%s %s &" edname fname
              and wd = Sys.getcwd () in
              Sys.chdir (Filename.dirname fname);
              let code = Sys.command cmd in
              if code = 0 then ()
              else app#warn ("Failed to execute " ^ cmd);
              Sys.chdir wd )
          | _ -> app#warn "No editor available for this file type" )
  
    method first_time_greeting () =
      (* XXX Maybe use GUtil.next_callback_id here? XXX *)
      let cont_id:GtkSignal.id option ref = ref None in
      let continue einf =
        self#clear ();
        ( match !cont_id with
          | None -> ()
          | Some id -> textwin#misc#disconnect id );
        app#show_help ();
        false in
      cont_id := Some (textwin#event#connect#button_release ~callback:continue);
      let datadir =
        try Sys.getenv "BANTAM_DATA"
        with Not_found -> failwith missing_vars_warning in
      let message =
        "\n\nThank you for using Bantam 0.4! This project has a very simple " ^
        "objective: to create the file manager of choice for Linux power " ^
        "users.\n\n" ^
        "Bantam is operated mainly by means of single-keystroke commands. " ^
        "Press '?' at any time to view a list of keystrokes and the commands " ^
        "they invoke.\n\n" ^
        "Further documentation may be found in " ^ datadir ^ "/doc/.\n\n" ^
        "Click anywhere in this panel to continue." in
      titlevar#set "~:| WELCOME |:~";
      let logofile = Filename.concat datadir "logo.png" in
      if Sys.file_exists logofile then
        ( let pb = GdkPixbuf.from_file logofile in
          let iter = textwin#buffer#get_iter `INSERT in 
          textwin#buffer#insert_pixbuf ~iter:iter ~pixbuf:pb;
          textwin#buffer#insert "\n" )
      else
        app#warn ("Logo image not found. This is not in itself a serious\n" ^
          "error, but it indicates that Bantam may not be correctly\n" ^
          "installed. Please check your installation.");
      textwin#buffer#insert message
  
end

class workbench app = object(self)
  val hu = GPack.hbox ()
  method hull = hu
  method set_focus () = ()
end


module MTBase = Weak.Make (
  struct
    type t = dirmodel_t
    let equal ms1 ms2 = (ms1#dir = ms2#dir)
    let hash ms = Hashtbl.hash ms#dir
  end
)
module MT = struct
  include MTBase
  let dump mt =
    print_endline "Model table:\n=============\n"; 
    Printf.printf "Size: %d\n\n" (count mt);
    print_endline "Models:\n--------";
    iter (fun mdl -> print_endline mdl#dir) mt;
    print_newline ();
    flush stdout
end


class dummy_model = object(self)
  val mutable dir_ = ""
  val vcol_ = GTree.view_column ()
  method coerce = (self :> dirmodel_t)
  method set_dir d = dir_ <- d
  method dir = dir_
  method mtime = 0.
  method vcol = vcol_
  method toggle_hidden () = ()
  method update () = ()
  method attach_to_panel p = ()
  method detach_from_panel p = ()
  method get_text _ = ""
end

class model_store = object(self)

  val dummy = new dummy_model
  val models = MT.create 23

  method get ?(must_exist=false) dir =
    dummy#set_dir dir;    
    try
      ( let dm = MT.find models dummy#coerce in
        if Util.get_mtime dir <> dm#mtime then dm#update ();
        dm )
    with Not_found ->
      ( if must_exist then raise Not_found
        else
          ( let dm = (new dir_model dir)#coerce in
            MT.add models dm;
            dm ) )
end

(* ================================================================== *)
(* --  MAIN APPLICATION CLASS  -------------------------------------- *)
class bantam =
    let mainwin_ = GWindow.window ~title:"Bantam" ~allow_grow:true
      ~allow_shrink:true ~border_width:4 () in
    let mainbox_ = GPack.hbox ~packing:mainwin_#add () in
    let dirbox_ = GPack.hbox ~packing:(mainbox_#pack ~expand:true) () in
    (* let dirbox_ = GPack.hbox ~packing:(mainbox_#pack ~expand:false) () in *)
    let tooltips_ = GData.tooltips ~delay:prefs#tooltip_delay () in
  object(self)
    val mstore = new model_store
    val mutable panels:dirpanel_t list = []
    val mutable viewpanel_:viewpanel option = None
    val mutable workbench_:workbench option = None
    val mutable last_id = 0
    val mutable viewpanel_visible = false
    val mutable guifont_ = GPango.font_description prefs#gui_font
    val mutable textfont_ = GPango.font_description prefs#text_font
    val mutable tipfont_ = GPango.font_description prefs#tip_font
    val mutable commands = []
    val keymap_ = BConfig.keymap ()
    val viewers_ = BConfig.viewers ()
    val editors_ = BConfig.editors ()
    (* val mutable help_tagcmds:(unit -> unit) list = [] *)
        (* Or should it be `INDENT ? *)
    val help_tagprops = ["label", [`LEFT_MARGIN 0]; "desc", [`LEFT_MARGIN 8]]
    val mutable help_text:(string * string) list = []
    val trash_root =
      let user_dir =
        try Sys.getenv "BANTAM_USER_DIR"
        with Not_found -> failwith missing_vars_warning in
      Filename.concat user_dir "trash"
    val mutable active:dirpanel_t option = None
    val mutable last_visited:dirpanel_t option = None
    val mutable defaults_ = { show_hidden = false }
    val mutable is_daemon = false
    val mutable msg_pipe = ""
    val mutable msg_fd = Unix.stdin
    val mutable setup_complete = false
    val idle_priority = Glib.int_of_priority `DEFAULT_IDLE
    val low_priority = Glib.int_of_priority `LOW
  
    initializer
      let vwr = new viewpanel (self :> app_t)
      and wkb = new workbench (self :> app_t) in
      viewpanel_ <- Some vwr;
      workbench_ <- Some wkb;
        (* Or should it be `INDENT ? *)
      (*
      help_tagcmds <-
        [(fun () -> self#viewpanel#tag_configure "label" [`LEFT_MARGIN 0]);
         (fun () -> self#viewpanel#tag_configure "desc" [`LEFT_MARGIN 8])];
      *)
      self#set_commands ();
      self#create_helptext ();
      self#set_bindings ()
  
    method set_commands () =
      commands <- [
        "default_action", (self#default_action,
        "Perform a default action according to the type of the selected " ^
        "file. Currently the only default action supported is to change " ^
        "to a directory. Future versions will support user-configurable " ^
        "default actions.");
        "close_viewpanel", (self#close_viewpanel,
        "If the text viewer is showing, close it.");
        "copy_file", (self#copy_file,
        "Copy the selected file(s) to another directory. Normally, the file "^
        "is copied to the directory shown in the next panel to the right. "^
        "If the starting point is the rightmost panel, the file will be  "^
        "copied to the next panel to the left. If only one directory is "^
        "visible, the operation is canceled.");
        "copy_file_menu", (self#copy_file_menu, "");
        "new_directory", (self#new_directory,
        "Prompt the user for a name, then create a directory with that name. " ^
        "Press <Return> to execute, or <Escape> to cancel.");
        "edit_file", (self#edit_file,
        "Load the selected file(s) in an external editor. Editors are defined " ^
        "on a per-file-type basis in the \"editors\" configuration file. If no " ^
        "editor is defined for the given file's type, the operation is " ^
        "canceled.");
        "edit_file_menu", (self#edit_file_menu, "");
        "goto_dir", (self#goto_dir,
        "Prompt the user to enter a directory path, then change to that " ^
        "directory in the current panel. Press <Return> to execute, or " ^
        "<Escape> to cancel.");
        "goto_dir_menu", (self#goto_dir_menu, "");
        "move_left", (self#move_left, "");
        "move_first", (self#move_first, "");
        "file_info", (self#file_info, "");
        "file_info_menu", (self#file_info_menu, "");
        "move_down", (self#move_down, "");
        "move_bottom", (self#move_bottom, "");
        "move_up", (self#move_up, "");
        "move_top", (self#move_top, "");
        "move_right", (self#move_right, "");
        "move_last", (self#move_last, "");
        "move_next", (self#move_next,
        "Move to the next panel, or if focus is in the last directory panel " ^
        "(or the viewpanel, if visible), wrap around to the first panel.");
        "move_prev", (self#move_prev,
        "Move to the previous panel, or if focus is in the first panel, " ^
        "wrap around to the first panel (or the viewpanel, if visible).");
        "move_file", (self#move_file,
        "Move the selected file(s) to another directory. Normally, the file " ^
        "is moved to the directory shown in the next panel to the right. " ^
        "If the starting point is the rightmost panel, the file will be  " ^
        "moved to the next panel to the left. If only one directory is " ^
        "visible, the operation is canceled.");
        "move_file_menu", (self#move_file_menu, "");
        "new_file", (self#new_file,
        "Prompt the user for a name, then create a file with that name. " ^
        "Press <Return> to execute, or <Escape> to cancel.");
        "trash_file", (self#trash_file,
        "Move the selected file(s) to the trash directory--normally, " ^
        "~/.bantam/trash. The full path is preserved in case you need " ^
        "to restore the file. Eventually there will be a routine for " ^
        "periodically emptying the trash directory, but that does not " ^
        "exist yet, so you will need to perform that task manually.");
        "trash_file_menu", (self#trash_file_menu, "");
        "view_file", (self#view_file,
        "Load the selected file(s) in an internal or external viewer. Viewers " ^
        "are defined on a per-file-type basis in the \"viewers\" configuration " ^
        "file. Currently, the internal viewer supports only text files, so you " ^
        "should define external viewers for all other file types. If no viewer " ^
        "is defined for the given file's type, the operation is canceled.");
        "view_file_menu", (self#view_file_menu, "");
        "add_to_workbench", (self#add_to_workbench, "");
        "add_to_workbench_menu", (self#add_to_workbench_menu, "");
        "add_panel", (self#add_panel,
        "Open a new directory panel on the right side of the window. If the " ^
        "currently selected file is a directory, the new panel will display " ^
        "that directory; otherwise, it will display the same directory as the " ^
        "current panel.");
        "add_and_goto_panel", (self#add_and_goto_panel,
        "Open a new directory panel, as in 'add-panel', and move to that panel.");
        "close_panel", (self#close_panel,
        "Close the current directory panel. If the current panel is the only " ^
        "one showing, exit the program.");
        "close_panel_menu", (self#close_panel_menu, "");
        "goto_workbench", (self#goto_workbench, "");
        "x_terminal_here", (self#x_terminal_here,
        "Launch an X terminal in the current directory.");
      ]
  
    method create_helptext () =
      let compare_keys ((sym1, sh1), _) ((sym2, sh2), _) =
        if sym1 = sym2 && sh2 && (not sh1) then -1
        else if sym1 = sym2 && sh1 && (not sh2) then 1 
        else if String.uppercase sym1 = sym2 then -1
        else if String.uppercase sym2 = sym1 then 1
        else compare sym1 sym2 in
      let rec build_txts kmap texts =
        match kmap with
        | [] -> List.rev texts
        | ((sym, shift), cmdname)::rest ->
          ( let _, desc = self#command cmdname in
            match desc with
            | "" -> build_txts rest texts
            | d -> 
              let key0 = Util.string_of_keysym sym in
              let key =
                if shift then "shift-" ^ key0
                else key0 in
              let lbl = "[" ^ key ^ "] " ^ cmdname ^ "\n" in
              build_txts rest
                (("desc", (d ^ "\n\n"))::("label", lbl)::texts) ) in
      (* FIXME: need a different comparison function *)
      let htexts = build_txts (List.sort compare keymap_) [] in
      help_text <- htexts
  
    method set_bindings () =
      ignore (mainwin_#event#connect#key_release ~callback:self#key_handler);
      ignore (mainwin_#event#connect#delete
        ~callback:(fun _ -> self#quit (); true))
  
    (* Widget access *)
    method w = mainwin_
    method p = dirbox_
    method viewpanel =
      match viewpanel_ with
      | None -> raise (IllegalState "Viewer has not been initialized.")
      | Some vp -> vp
    method workbench =
      match workbench_ with
      | None -> raise (IllegalState "Workbench has not been initialized.")
      | Some wb -> wb
    method tooltips = tooltips_
    method guifont = guifont_
    method textfont = textfont_
    method tipfont = tipfont_
    method defaults = defaults_

    method private height =
      let alloc = self#w#misc#allocation in
      alloc.Gtk.height
    method private set_size () =
      let vpw =
        let w, _ =
          if viewpanel_visible then self#viewpanel#size
          else 0, 0 in
        w in
      let width =
        List.fold_left
          ( fun w1 p -> let w2, _ = p#size in w1 + w2 + 1 )
          (* 8 is border width * 2 *)
          (vpw+8) panels in
      let width' = min width (Gdk.Screen.width ())
      and height' =
        if setup_complete then self#height
        else min prefs#panel_height (Gdk.Screen.height ()) in
      self#w#resize width' height'
  
    method private key_handler ev =
      if GdkEvent.Key.keyval ev = _question then (self#show_help (); true)
      else false
    method command cmdname = List.assoc cmdname commands
    method keymap = keymap_
    method private handler handler_map file =
      prerr_endline "#bantam#handler";
      let mtype = Util.default_media_type file in
      let result =
      ( match mtype with
      | Text _ ->
        ( prerr_endline "Text";
          try List.assoc mtype handler_map
          with Not_found ->
            ( prerr_endline "OtherText";
              try List.assoc OtherText handler_map
              with Not_found -> List.assoc Other handler_map ) )
      | Image _ ->
        ( prerr_endline "Image";
          try List.assoc mtype handler_map
          with Not_found ->
            ( prerr_endline "OtherImage";
              try List.assoc OtherImage handler_map 
              with Not_found -> List.assoc Other handler_map ) )
      | Audio _ ->
        ( prerr_endline "Audio";
          try List.assoc mtype handler_map
          with Not_found ->
            ( prerr_endline "OtherAudio";
              try List.assoc OtherAudio handler_map 
              with Not_found -> List.assoc Other handler_map ) )
      | Video _ ->
        ( prerr_endline "Video";
          try List.assoc mtype handler_map
          with Not_found ->
            ( prerr_endline "OtherVideo";
              try List.assoc OtherVideo handler_map 
              with Not_found -> List.assoc Other handler_map ) )
      | Application _ ->
        ( prerr_endline "Application";
          try List.assoc mtype handler_map
          with Not_found ->
            ( prerr_endline "OtherApplication";
              try List.assoc OtherApplication handler_map 
              with Not_found -> List.assoc Other handler_map ) )
      | OtherText ->
        ( prerr_endline "OtherText";
          try List.assoc mtype handler_map
          with Not_found -> List.assoc Other handler_map )
      | OtherImage ->
        ( prerr_endline "OtherImage";
          try List.assoc mtype handler_map
          with Not_found -> List.assoc Other handler_map )
      | OtherAudio ->
        ( prerr_endline "OtherAudio";
          try List.assoc mtype handler_map
          with Not_found -> List.assoc Other handler_map )
      | OtherVideo ->
        ( prerr_endline "OtherVideo";
          try List.assoc mtype handler_map
          with Not_found -> List.assoc Other handler_map )
      | OtherApplication ->
        ( prerr_endline "OtherApplication";
          try List.assoc mtype handler_map
          with Not_found -> List.assoc Other handler_map )
      | Custom _ ->
        ( prerr_endline "Custom";
          try List.assoc mtype handler_map
          with Not_found -> List.assoc Other handler_map )
      (* | Other -> List.assoc Other handler_map *)
      | Other ->
        ( prerr_endline "Other";
          List.assoc Other handler_map ) ) in
      flush stderr;
      result
    method viewer file = self#handler viewers_ file
    method editor file = self#handler editors_ file
  
    (* Utility methods *)
    method warn = Util.warn

    method get_model dir =
      mstore#get dir
  
    method private new_panel_id =
      let next = last_id + 1 in
      last_id <- next;
      next
  
    method int_add_panel ?dir ?(go=false) () =
      let dir =
        match dir with
        | None -> Sys.getcwd ()
        | Some d -> d in
      let id = self#new_panel_id in
      let p = new dirpanel (self :> app_t) id in
      panels <- panels @ [p];
      self#set_size ();
      p#set_directory dir;
      if go then ignore (p#set_focus ())
  
    method update_views dir =
      try let dm = mstore#get ~must_exist:true dir in dm#update ()
      with Not_found -> self#warn ("No model found for " ^ dir)
  
    method next_panel panel =
      let rec search pnls =
        match pnls with
        | [] -> None
        | [p] -> None
        | p1::p2::_ when p1 == panel -> Some p2
        | _::pp -> search pp in
      search panels
    method prev_panel panel =
      let rec search ?last pnls =
        match pnls with
        | [] -> None
        | p::_ when p == panel -> last
        | p::pp -> search ~last:p pp in
      search panels
  
    method set_active panel (ev:GdkEvent.Focus.t) = active <- Some panel; true
    method private int_unset_active () = active <- None
    method unset_active (ev:GdkEvent.Focus.t) = active <- None; true
    method set_last panel (ev:GdkEvent.Focus.t) =
      last_visited <- Some panel; true
  
    method private first_panel = List.hd panels
    method private last_panel = List.hd (List.rev panels)
  
    method private show_viewpanel () =
      if viewpanel_visible then ()
      else
        ( mainbox_#pack ~expand:true self#viewpanel#hull#coerce;
          ignore (self#viewpanel#set_focus ());
          viewpanel_visible <- true;
          self#set_size () )
    method private hide_viewpanel () =
      if not viewpanel_visible then ()
      else
        ( mainbox_#remove self#viewpanel#hull#coerce;
          viewpanel_visible <- false;
          self#set_size () )
  
    method int_view_file fname =
      self#show_viewpanel ();
      self#viewpanel#clear (); 
      self#viewpanel#load_file fname
      
    method int_view_text title text =
      self#show_viewpanel ();
      self#viewpanel#clear (); 
      self#viewpanel#load_text title text
  
    method int_view_tagtext title tagcmds txt =
      self#show_viewpanel ();
      self#viewpanel#clear (); 
      self#viewpanel#load_tagtext title tagcmds txt
  
    method show_help () =
      self#int_view_tagtext "[ HELP ]" help_tagprops help_text
  
    (* This will generally be called from the viewpanel *)
    method int_close_viewpanel einf =
      self#hide_viewpanel ();
      match last_visited with
      | Some p -> ignore (p#set_focus ())
      | None -> ignore (self#last_panel#set_focus ())
    method close_viewpanel panel =
      self#hide_viewpanel ()
  
(*     method copy_file panel einf = *)
    method copy_file panel  =
      let file = panel#selected_file in
      match file with
      | NoSelection -> ()
      | ParentDir -> ()
      | File f ->
        ( let to_panel =
            match (self#next_panel panel) with
            | Some p -> Some p
            | None -> self#prev_panel panel in
          match to_panel with
          | Some p ->
            if Util.has_same_file f p#directory then
              ( let proposed_name = Util.newname f p#directory in
                let idx =
                  try Util.basename_start_idx proposed_name
                  with Not_found -> 0 in
                let cb newname = p#copy_to ~name:newname f in
                panel#do_with_param ~default:proposed_name
                  ~start:idx "Copy as" cb )
            else p#copy_to f
          | None -> () )
(*     method copy_file_menu panel einf = *)
    method copy_file_menu panel  =
      print_endline "copy_file_menu"
    method new_directory panel =
      let cb newname =
        match newname with
        | "" -> ()
        | nn -> Util.create_directory panel#directory nn in
      panel#do_with_param "Directory name" cb;
      self#update_views panel#directory
(*     method edit_file panel einf = *)
    method edit_file panel =
      let file = panel#selected_file in
      match file with
      | NoSelection -> ()
      | ParentDir -> ()
      | File f ->
        ( let ed = self#editor f in
          match ed with
          | ExtHandler edname ->
            ( let cmd = Printf.sprintf "%s %s" edname f
              and wd = Sys.getcwd () in
              Sys.chdir panel#directory;
              let code = Sys.command cmd in
              if code = 0 then ()
              else self#warn ("Failed to execute " ^ cmd);
              Sys.chdir wd )
          | _ ->
            self#warn "No editor available for this file type." )
(*     method edit_file_menu panel einf = *)
    method edit_file_menu panel  =
      print_endline "edit_file_menu"
  
    (* FIXME: Need one function for return, and another for 'g' *)
(*     method goto_dir panel einf = *)
    method goto_dir panel =
      let curdir = panel#directory in
      let cb proposed =
        if proposed <> curdir then
          ( try
              let newdir = Util.normdir curdir proposed in
              panel#save_selected ();
              panel#set_directory newdir;
              panel#set_selected ()
            with _ -> self#warn ("Can't change to " ^ proposed) ) in
      panel#do_with_param "Go to" cb
  
(*     method default_action panel einf = *)
    method default_action panel  =
      let curdir = panel#directory
      and cd_handler dir =
        try panel#set_directory dir
        with _ -> self#warn ("Can't change to " ^ dir ^ ".") in
      let action =
        match panel#selected_file with
        | NoSelection -> None
        | ParentDir ->
          Some (fun () -> cd_handler (Util.parent curdir))
        | File f ->
          ( match Util.get_filetype f with
            | Directory | Dirlink ->
              let newdir = Util.normdir curdir f in
              Some (fun () -> cd_handler newdir)
            | _ -> None ) in
      match action with
      | None -> ()
      | Some a ->
        ( panel#save_selected ();
          a ();
          panel#set_selected () )
  
(*     method goto_dir_menu panel einf = *)
    method goto_dir_menu panel =
      print_endline "goto_dir_menu"
(*     method move_left panel einf = *)
    method move_left panel =
      print_endline "move_left"
    method move_first (panel:dirpanel_t) =
      print_endline "move_first"
(*     method file_info panel einf = *)
    method file_info panel =
      print_endline "file_info"
(*     method file_info_menu panel einf = *)
    method file_info_menu panel =
      print_endline "file_info_menu"
(*     method move_down panel einf = *)
    method move_down panel =
      print_endline "move_down"
(*     method move_bottom panel einf = *)
    method move_bottom panel =
      print_endline "move_bottom"
(*     method move_up panel einf = *)
    method move_up panel =
      print_endline "move_up"
(*     method move_top panel einf = *)
    method move_top panel =
      print_endline "move_top"
(*     method move_right panel einf = *)
    method move_right panel =
      print_endline "move_right"
    method move_last (panel:dirpanel_t) =
      print_endline "move_last"
    method move_next panel =
      let next = self#next_panel panel in
      match next with
      | Some p -> ignore (p#set_focus ())
      | None when viewpanel_visible -> ignore (self#viewpanel#set_focus ())
      | None -> ignore (self#first_panel#set_focus ())
    method move_prev panel =
      let prev = self#prev_panel panel in
      match prev with
      | Some p -> ignore (p#set_focus ())
      | None when viewpanel_visible -> ignore (self#viewpanel#set_focus ())
      | None -> ignore (self#last_panel#set_focus ())
(*     method move_file panel einf = *)
    method move_file panel =
      let file = panel#selected_file in
      match file with
      | NoSelection -> ()
      | ParentDir -> ()
      | File f ->
        ( let to_panel =
            match (self#next_panel panel) with
            | Some p -> Some p
            | None -> self#prev_panel panel in
          match to_panel with
          | Some p ->
            if Util.has_same_file f p#directory then
              ( let proposed_name = Util.newname f p#directory in
                let idx =
                  try Util.basename_start_idx proposed_name
                  with Not_found -> 0 in
                let cb newname =
                  let moved = p#move_to ~name:newname f in
                  if moved && (p#directory <> panel#directory) then
                    ( panel#update ();
                      panel#set_selected () ) in
                panel#do_with_param ~default:proposed_name ~start:idx
                  "Move as" cb )
            else if p#move_to f then ( panel#update (); panel#set_selected () )
            else ()
          | None -> () )
(*     method move_file_menu panel einf = *)
    method move_file_menu panel =
      print_endline "move_file_menu"
    method new_file panel =
      let cb newname =
        match newname with
        | "" -> ()
        | nn -> Util.create_file panel#directory nn in
      panel#do_with_param "File name" cb;
      self#update_views panel#directory
(*     method trash_file panel einf = *)
    method trash_file panel =
      let file = panel#selected_file in
      match file with
      | NoSelection -> ()
      | ParentDir -> ()
      | File f ->
        ( let trash_dir = Filename.concat trash_root panel#directory in
          let cmd = "mkdir -p " ^ trash_dir in
          let code = Sys.command cmd in 
          if code = 0 then 
            ( let cmd = (Printf.sprintf "mv %s %s" f trash_dir) in
              let code = Sys.command cmd in
              if code = 0 then 
                ( self#update_views panel#directory;
                  panel#set_selected () )
              else self#warn
                ("Failed to move file to trash directory " ^ trash_dir) )
          else
            self#warn ("Failed to create trash directory " ^ trash_dir) )
(*     method trash_file_menu panel einf = *)
    method trash_file_menu panel =
      print_endline "trash_file_menu"
(*     method view_file panel einf = *)
    method view_file panel =
      let file = panel#selected_file in
      match file with
      | NoSelection -> ()
      | ParentDir -> ()
      | File f ->
        ( let vwr = self#viewer f in
          match vwr with
          | ExtHandler vname ->
            ( let cmd = Printf.sprintf "%s %s &" vname f
              and wd = Sys.getcwd () in
              Sys.chdir panel#directory;
              let code = Sys.command cmd in
              if code = 0 then ()
              else self#warn ("Failed to execute " ^ cmd);
              Sys.chdir wd )
          | IntHandler -> self#int_view_file f
          | NoHandler ->
            self#warn "No viewer available for this file type." )
(*     method view_file_menu panel einf = *)
    method view_file_menu panel =
      print_endline "view_file_menu"
(*     method add_to_workbench panel einf = *)
    method add_to_workbench panel =
      print_endline "add_to_workbench"
(*     method add_to_workbench_menu panel einf = *)
    method add_to_workbench_menu panel =
      print_endline "add_to_workbench_menu"
  
    method private add_panel_ ?(go=false) (panel:dirpanel_t) =
      let curdir = panel#directory
      and newdir0 =
        match panel#selected_file with
        | NoSelection -> None
        | ParentDir -> Some (Util.parent panel#directory)
        | File f ->
          ( match Util.get_filetype f with
            | Directory | Dirlink -> Some f
            | _ -> None ) in
      let cb proposed =
        let newdir = Util.normdir curdir proposed in
        self#int_add_panel ~dir:newdir ~go:go () in
      match newdir0 with
      | Some d -> self#int_add_panel ~dir:d ~go:go ()
      | None -> panel#do_with_param ~default:"." "Go to" cb
(*     method add_panel panel einf = *)
    method add_panel panel =
      self#add_panel_ panel
(*     method add_and_goto_panel panel einf = *)
    method add_and_goto_panel panel =
      self#add_panel_ ~go:true panel
  
(*     method close_panel panel einf = *)
    method close_panel panel =
      if List.length panels < 2 then self#quit ()
      else
        ( let dest =
            match last_visited with
            | Some p when p = panel -> ( last_visited <- None; None )
            | lv -> lv in
          let dest' =
            match dest with
            | Some p -> Some p
            | None ->
              ( match self#prev_panel panel with
                | Some p -> Some p
                | None -> self#next_panel panel ) in
          let remaining =
            List.filter ( fun p -> p != panel ) panels in
          panels <- remaining;
          panel#destroy ();
          self#int_unset_active ();
          ( match dest' with
            | Some p -> ignore (p#set_focus ())
            (* | None -> () ); *)
            | None -> (prerr_endline "no dest??"; flush stderr) );
          self#set_size () )
(*     method close_panel_menu panel einf = *)
    method close_panel_menu panel =
      print_endline "close_panel_menu"
(*     method goto_workbench panel einf = *)
    method goto_workbench panel =
      print_endline "goto_workbench"
  
(*     method x_terminal_here panel einf = *)
    method x_terminal_here panel =
      try
        (* let xterm_command = BConfig.getpref "x_terminal" in *)
        (* Util.launch ~dir:panel#directory xterm_command *)
        Util.launch ~dir:panel#directory prefs#x_terminal
      with Not_found -> ()

    method first_time_setup () =
      self#int_add_panel ();
      self#first_time_greeting ();
      self#show ()

    method daemon_setup () =
      self#int_add_panel ~go:true ();
      is_daemon <- true;
      msg_pipe <- Filename.concat Util.temp_user_dir "message.pipe";
      if not (Sys.file_exists Util.temp_user_dir) then
        Unix.mkdir Util.temp_user_dir 0o755
      else if Sys.file_exists msg_pipe then
        Unix.unlink msg_pipe;
      Unix.mkfifo msg_pipe 0o600;
      msg_fd <-
        (* Unix.openfile msg_pipe [Unix.O_RDONLY; Unix.O_NONBLOCK] 0o600; *)
        Unix.openfile msg_pipe [Unix.O_RDONLY] 0o600;
      let chan = GMain.Io.channel_of_descr msg_fd in
      let input_watcher _ =
        let cmdbuf = Buffer.create 4
        and temp = " " in
        let rec read_input ready =
          let nchars =
            try Some (GMain.Io.read chan ~buf:temp ~pos:0 ~len:1)
            with Glib.GError "g_io_channel_read: G_IO_ERROR_AGAIN" -> None in
          match nchars with
          | Some 0 -> Util.sleep 0.25
          | Some 1 ->
            ( match temp with
              | "%" -> read_input true
              | "\n" -> read_input false
              | s when ready -> (Buffer.add_string cmdbuf s; read_input true)
              | s -> (self#warn ("Unexpected input: " ^ s); read_input false) )
          | Some _ -> failwith "Wrong # of input characters??"
          (* | None -> read_input ready *)
          | None -> Util.sleep 0.25
        and do_command cmdstring =
          match cmdstring with
          | "show" -> self#show ()
          | "hide" -> self#hide ()
          | "quit" -> self#really_quit ()
          | "hup" -> self#hup ()
          | cs -> self#warn ("Unknown command: " ^ cs) in
        read_input false; 
        do_command (Buffer.contents cmdbuf);
        true
      and hup_watcher _ = (Util.sleep 0.05; true) in
      (* pa_macro doesn't like `END, so we'll use sed for now. *)
      ignore (GMain.Io.add_watch ~prio:idle_priority ~cond:@IN_COND@
        ~callback:input_watcher chan);
      ignore (GMain.Io.add_watch ~prio:low_priority ~cond:@HUP_COND@
        ~callback:hup_watcher chan)
      (*
      IFDEF NEW_ADD_WATCH THEN
      ignore (GMain.Io.add_watch ~cond:[`IN] ~callback:watcher chan)
      ELSE
      ignore (GMain.Io.add_watch ~cond:`IN ~callback:watcher chan)
      ENDIF
      *)
      
    method normal_setup () =
      self#int_add_panel ~go:true ();
      self#show ()
  
    method show () = mainwin_#misc#show_all ()
    method hide () = mainwin_#misc#hide ()

    method private hup () = ()

    method private really_quit () =
      if is_daemon then ( Unix.close msg_fd; Unix.unlink msg_pipe );
      try (mainwin_#destroy (); exit 0)
      with _ -> exit 1

    method quit () =
      if is_daemon then self#hide ()
      else self#really_quit ()
  
    method first_time_greeting () =
      self#show_viewpanel ();
      self#viewpanel#first_time_greeting ()

    method init ?(first=false) ?(daemon=false) () =
      if first then self#first_time_setup ()
      else if daemon then self#daemon_setup ()
      else self#normal_setup ();
      setup_complete <- true;
      GMain.Main.main ()
end
(* ================================================================== *)


let run () =
  let b = new bantam in
  b#init ()

let first_run () =
  let b = new bantam in
  b#init ~first:true ()

let run_daemon () =
  let b = new bantam in 
  b#init ~daemon:true ()

let _ =
  if Array.length Sys.argv > 1 && Sys.argv.(1) = "-first" then
    first_run ()
  else if Array.length Sys.argv > 1 && Sys.argv.(1) = "-daemon" then
    run_daemon ()
  else
    run ()

let _ =
  if Array.length Sys.argv > 1 && Sys.argv.(1) = "-first" then
    first_run ()
  else if Array.length Sys.argv > 1 && Sys.argv.(1) = "-daemon" then
    run_daemon ()
  else
    run ()
