(* Bantam -- a lightweight file manager for X11
   Copyright  2005 by Matthew C. Gushee

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)

open Common
open GdkKeysyms


let temp_user_dir =
  let tmp =
    try Sys.getenv "TMPDIR"
    with _ -> "/tmp"
  and user = Sys.getenv "USER" in
  if Sys.file_exists tmp then Filename.concat tmp ("bantam-" ^ user)
  else failwith ("Temporary directory not found. Try setting " ^
                 "the TMPDIR environment variable.")

let warn msg =
  prerr_endline ("WARNING: " ^ msg);
  flush stderr
  

let round_int x =
  let f, i = modf x in
  if f >= 0.5 then int_of_float i + 1
  else int_of_float i

let get_mtime dir =
  (Unix.stat dir).Unix.st_mtime

let get_filetype file =
  try
    ( let lstat = Unix.lstat file in
      match lstat.Unix.st_kind with
      | Unix.S_REG -> Regular
      | Unix.S_DIR -> Directory
      | Unix.S_CHR -> Device
      | Unix.S_BLK -> Device
      | Unix.S_LNK ->
        ( try
            let stat = Unix.stat file in
            match stat.Unix.st_kind with
            | Unix.S_DIR -> Dirlink
            | _ -> Symlink
          with _ ->
            ( prerr_endline
              (Printf.sprintf "WARNING: stat %s failed." file);
              Unknown ) )
      | Unix.S_FIFO -> FIFO
      | Unix.S_SOCK -> Socket )
   with _ ->
     ( prerr_endline (Printf.sprintf "lstat %s failed" file);
       Unknown )

let mark_filetype file =
  match (get_filetype file) with
  | Regular -> file ^ " "
  | Symlink -> file ^ "@"
  | Directory -> file ^ "/"
  | Dirlink -> file ^ "%"
  | Device -> file ^ ":"
  | FIFO -> file ^ "|"
  | Socket -> file ^ "="
  | Unknown -> file ^ "?"

let list_files ?(show_hidden=false) dir =
  let startdir = Sys.getcwd () in
  Unix.chdir dir;
  let actual_dir = Sys.getcwd () in
  let dh = Unix.opendir "." in
  let mtime = get_mtime "." in
  let show_file f =
    match f, show_hidden with
    | ".", _ -> false
    | "..", _ -> false
    | f', false -> not (f'.[0] = '.')
    | _, true -> true in
  let rec read_list files =
    let next =
      try Some (Unix.readdir dh)
      with End_of_file -> None in
    match next with
    | Some f when (show_file f) -> read_list ((mark_filetype f)::files)
    | Some f -> read_list files
    | None when actual_dir = "/" -> List.sort compare files
      (* The trailing space is intentional--every filename gets one
       * character appended to indicate its type. *)
    | None -> ".. "::(List.sort compare files) in
  let ls = read_list [] in
  Unix.closedir dh;
  Unix.chdir startdir;
  ls, mtime

let create_file parent fname =
  let startdir = Sys.getcwd () in
  Unix.chdir parent;
  ( if Sys.file_exists fname then ()
    else ignore (Sys.command ("touch " ^ fname)) );
  Unix.chdir startdir

let create_directory parent dirname =
  let startdir = Sys.getcwd () in
  Unix.chdir parent;
  ( if Sys.file_exists dirname then ()
    else Unix.mkdir dirname 0o755 ); 
  Unix.chdir startdir

let newname fname dir =
  Filename.concat dir (Filename.basename fname)

let has_same_file fname dir =
  Sys.file_exists (newname fname dir)

let basename_start_idx fname =
  let dname = Filename.dirname fname in
  if dname = fname then raise Not_found;
  String.length dname + 1

let chop_dir dir width =
  let len = String.length dir in
  if len > width then
    let start = len - width in
    "..." ^ (String.sub dir start width)
  else dir

let normdir olddir newdir =
  let wd = Sys.getcwd () in
  Sys.chdir olddir;
  Sys.chdir newdir; 
  let result = Sys.getcwd () in
  Sys.chdir wd;
  result

let parent dir =
  let wd = Sys.getcwd () in
  Sys.chdir dir;
  Sys.chdir "..";
  let result = Sys.getcwd () in
  Sys.chdir wd;
  result

let read_file fname =
  let ic = open_in fname in
  let rec readlines ?(init=false) txt =
    let line =
      try Some (input_line ic)
      with End_of_file -> None in
    match line with
    | None -> txt
    | Some ln when init -> readlines ln
    | Some ln -> readlines (txt ^ "\n" ^ ln) in
  let result = readlines ~init:true "" in
  close_in ic;
  Glib.Convert.locale_to_utf8 result 

let match_fullname names mtype filename =
  if (List.mem filename names) then Some mtype
  else None

let match_fullname_ci names mtype filename =
  let lnames = List.map String.lowercase names in
  if (List.mem (String.lowercase filename) lnames) then Some mtype
  else None

let match_basename names mtype filename =
  let base = Filename.basename filename in
  if (List.mem base names) then Some mtype
  else None

let match_basename_ci names mtype filename =
  let lnames = List.map String.lowercase names in
  let base = String.lowercase (Filename.basename filename) in
  if (List.mem base lnames) then Some mtype
  else None

let match_ext exts mtype filename =
  let lexts = List.map String.lowercase exts
  and lfile = String.lowercase filename in
  if (List.exists (Filename.check_suffix lfile) lexts) then Some mtype
  else None

let file_magic_contains strings mtype filename =
  let patt =
    match strings with
    | [] -> failwith "Nothing to match."
    | [s] -> s
    | s::ss ->
      let body = 
        (List.fold_left (fun s1 s2 -> s1 ^ "\\|" ^ s2) s ss) in
      "\\(" ^ body ^ "\\)" in
  let re = Str.regexp patt in
  let ic = Unix.open_process_in ("file " ^ filename) in
  try
    let magic = input_line ic in
    close_in ic;
    ( try ignore (Str.search_forward re magic 0); Some mtype
      with Not_found -> None )
  with _ -> None

let fallback mtype filename = Some mtype

(* Probably should be configurable *)
let default_media_type:(string -> media_type) =
  fun filename ->
    let mtype_tests = [
      (match_basename_ci ["readme"; "install"; "copying"; "license";
                          "bugs"; "news"; "changes"; "history";
                          "changelog"; "makefile"] (Text "plain"));
      (match_ext ["txt"; ".ml"; "mli"; "mll"; ".q"; ".py"; ".rb";
                  "lisp"; "scm"; ".c"; ".cxx"; ".h"; ".pl"; ".tcl";
                  ".java"] (Text "plain"));
      (match_ext ["htm"; "html"] (Text "html"));
      (match_ext ["xml"; "xsl"; "xslt"] (Application "xml"));
      (match_ext ["rtf"] (Text "rtf"));
      (match_ext ["css"] (Text "css"));
      (match_ext [".js"] (Text "javascript"));
      (match_ext ["jpg"; "jpeg"; "jpe"] (Image "jpeg"));
      (match_ext ["gif"] (Image "gif"));
      (match_ext ["png"] (Image "png"));
      (match_ext ["bmp"] (Image "bmp"));
      (match_ext ["xbm"] (Image "xbm"));
      (match_ext ["xpm"] (Image "xpm"));
      (match_ext ["pbm"] (Image "pbm"));
      (match_ext ["pgm"] (Image "pgm"));
      (match_ext ["ppm"] (Image "ppm"));
      (match_ext ["pnm"] (Image "pnm"));
      (match_ext ["tif"; "tiff"] (Image "tiff"));
      (match_ext ["mp3"; "mp2"; "mpga"] (Audio "mpeg"));
      (match_ext ["aif"; "aiff"; "aifc"] (Audio "x-aiff"));
      (match_ext ["wav"] (Audio "x-wav"));
      (match_ext ["mpeg"; "mpg"; "mpe"] (Video "mpeg"));
      (match_ext ["avi"] (Video "x-msvideo"));
      (match_ext ["pdf"] (Application "pdf"));
      (match_ext ["gz"] (Application "x-gzip"));
      (match_ext ["zip"] (Application "zip"));
      (match_ext [".ps"; ".eps"; ".ai"] (Application "postscript"));
      (match_ext ["tar"] (Application "x-tar"));
      (match_ext ["tex"] (Application "x-tex"));
      (match_ext ["latex"; "ltx"] (Application "x-latex"));
      (file_magic_contains ["text"] (OtherText));
      (file_magic_contains ["image"] (OtherImage));
      (file_magic_contains ["ELF"] (Application "octet-stream"));
      (fallback Other)
    ] in
    let rec run_tests tests =
      match tests with
      | [] -> failwith "Media type tests failed. Was a fallback defined?"
      | t::tt ->
        ( match t filename with
          | Some mtype -> mtype
          | None -> run_tests tt ) in
    run_tests mtype_tests

let launch ?dir ?(bg=true) command =
  let wd = Sys.getcwd () in
  ( match dir with
    | None -> ()
    | Some d -> Sys.chdir d );
  let cmd =
    if bg then command ^ " &"
    else command in
  let result, exc =
    try Sys.command cmd, None
    with e -> 1, Some e in
  ( match dir with
    | None -> ()
    | Some _ -> Sys.chdir wd );
  match result, exc with
  | 0, None -> ()
  | _, None -> failwith ("Failed to execute '" ^ cmd ^ "'")
  | _, Some e -> raise e


let keysym_of_string str =
  match str with
  | "voidsymbol" -> _VoidSymbol
  | "backspace" -> _BackSpace
  | "tab" -> _Tab
  | "left_tab" -> _ISO_Left_Tab
  | "linefeed" -> _Linefeed
  | "clear" -> _Clear
  | "return" -> _Return
  | "pause" -> _Pause
  | "scroll_lock" -> _Scroll_Lock
  | "sys_req" -> _Sys_Req
  | "escape" -> _Escape
  | "delete" -> _Delete
  | "multi_key" -> _Multi_key
  | "home" -> _Home
  | "left" -> _Left
  | "up" -> _Up
  | "right" -> _Right
  | "down" -> _Down
  | "prior" -> _Prior
  | "page_up" -> _Page_Up
  | "next" -> _Next
  | "page_down" -> _Page_Down
  | "end" -> _End
  | "begin" -> _Begin
  | "select" -> _Select
  | "print" -> _Print
  | "execute" -> _Execute
  | "insert" -> _Insert
  | "undo" -> _Undo
  | "redo" -> _Redo
  | "menu" -> _Menu
  | "find" -> _Find
  | "cancel" -> _Cancel
  | "help" -> _Help
  | "break" -> _Break
  | "mode_switch" -> _Mode_switch
  | "script_switch" -> _script_switch
  | "num_lock" -> _Num_Lock
  | "kp_space" -> _KP_Space
  | "kp_tab" -> _KP_Tab
  | "kp_enter" -> _KP_Enter
  | "kp_f1" -> _KP_F1
  | "kp_f2" -> _KP_F2
  | "kp_f3" -> _KP_F3
  | "kp_f4" -> _KP_F4
  | "kp_home" -> _KP_Home
  | "kp_left" -> _KP_Left
  | "kp_up" -> _KP_Up
  | "kp_right" -> _KP_Right
  | "kp_down" -> _KP_Down
  | "kp_prior" -> _KP_Prior
  | "kp_page_up" -> _KP_Page_Up
  | "kp_next" -> _KP_Next
  | "kp_page_down" -> _KP_Page_Down
  | "kp_end" -> _KP_End
  | "kp_begin" -> _KP_Begin
  | "kp_insert" -> _KP_Insert
  | "kp_delete" -> _KP_Delete
  | "kp_equal" -> _KP_Equal
  | "kp_multiply" -> _KP_Multiply
  | "kp_add" -> _KP_Add
  | "kp_separator" -> _KP_Separator
  | "kp_subtract" -> _KP_Subtract
  | "kp_decimal" -> _KP_Decimal
  | "kp_divide" -> _KP_Divide
  | "kp_0" -> _KP_0
  | "kp_1" -> _KP_1
  | "kp_2" -> _KP_2
  | "kp_3" -> _KP_3
  | "kp_4" -> _KP_4
  | "kp_5" -> _KP_5
  | "kp_6" -> _KP_6
  | "kp_7" -> _KP_7
  | "kp_8" -> _KP_8
  | "kp_9" -> _KP_9
  | "f1" -> _F1
  | "f2" -> _F2
  | "f3" -> _F3
  | "f4" -> _F4
  | "f5" -> _F5
  | "f6" -> _F6
  | "f7" -> _F7
  | "f8" -> _F8
  | "f9" -> _F9
  | "f10" -> _F10
  | "f11" -> _F11
  | "l1" -> _L1
  | "f12" -> _F12
  | "l2" -> _L2
  | "f13" -> _F13
  | "l3" -> _L3
  | "f14" -> _F14
  | "l4" -> _L4
  | "f15" -> _F15
  | "l5" -> _L5
  | "f16" -> _F16
  | "l6" -> _L6
  | "f17" -> _F17
  | "l7" -> _L7
  | "f18" -> _F18
  | "l8" -> _L8
  | "f19" -> _F19
  | "l9" -> _L9
  | "f20" -> _F20
  | "l10" -> _L10
  | "f21" -> _F21
  | "r1" -> _R1
  | "f22" -> _F22
  | "r2" -> _R2
  | "f23" -> _F23
  | "r3" -> _R3
  | "f24" -> _F24
  | "r4" -> _R4
  | "f25" -> _F25
  | "r5" -> _R5
  | "f26" -> _F26
  | "r6" -> _R6
  | "f27" -> _F27
  | "r7" -> _R7
  | "f28" -> _F28
  | "r8" -> _R8
  | "f29" -> _F29
  | "r9" -> _R9
  | "f30" -> _F30
  | "r10" -> _R10
  | "f31" -> _F31
  | "r11" -> _R11
  | "f32" -> _F32
  | "r12" -> _R12
  | "f33" -> _F33
  | "r13" -> _R13
  | "f34" -> _F34
  | "r14" -> _R14
  | "f35" -> _F35
  | "r15" -> _R15
  | "shift_l" -> _Shift_L
  | "shift_r" -> _Shift_R
  | "control_l" -> _Control_L
  | "control_r" -> _Control_R
  | "caps_lock" -> _Caps_Lock
  | "shift_lock" -> _Shift_Lock
  | "meta_l" -> _Meta_L
  | "meta_r" -> _Meta_R
  | "alt_l" -> _Alt_L
  | "alt_r" -> _Alt_R
  | "super_l" -> _Super_L
  | "super_r" -> _Super_R
  | "hyper_l" -> _Hyper_L
  | "hyper_r" -> _Hyper_R
  | "space" -> _space
  | "exclam" -> _exclam
  | "quotedbl" -> _quotedbl
  | "\"" -> _quotedbl
  | "numbersign" -> _numbersign
  | "#" -> _numbersign
  | "dollar" -> _dollar
  | "$" -> _dollar
  | "percent" -> _percent
  | "%" -> _percent
  | "ampersand" -> _ampersand
  | "&" -> _ampersand
  | "apostrophe" -> _apostrophe
  | "'" -> _apostrophe
  | "quoteright" -> _quoteright
  | "`" -> _quoteright
  | "parenleft" -> _parenleft
  | "(" -> _parenleft
  | "parenright" -> _parenright
  | ")" -> _parenright
  | "asterisk" -> _asterisk
  | "*" -> _asterisk
  | "plus" -> _plus
  | "+" -> _plus
  | "comma" -> _comma
  | "," -> _comma
  | "minus" -> _minus
  | "-" -> _minus
  | "period" -> _period
  | "." -> _period
  | "slash" -> _slash
  | "/" -> _slash
  | "0" -> _0
  | "1" -> _1
  | "2" -> _2
  | "3" -> _3
  | "4" -> _4
  | "5" -> _5
  | "6" -> _6
  | "7" -> _7
  | "8" -> _8
  | "9" -> _9
  | "colon" -> _colon
  | "semicolon" -> _semicolon
  | ";" -> _semicolon
  | "less" -> _less
  | "<" -> _less
  | "equal" -> _equal
  | "=" -> _equal
  | "greater" -> _greater
  | ">" -> _greater
  | "question" -> _question
  | "?" -> _question
  | "at" -> _at
  | "A" -> _A
  | "B" -> _B
  | "C" -> _C
  | "D" -> _D
  | "E" -> _E
  | "F" -> _F
  | "G" -> _G
  | "H" -> _H
  | "I" -> _I
  | "J" -> _J
  | "K" -> _K
  | "L" -> _L
  | "M" -> _M
  | "N" -> _N
  | "O" -> _O
  | "P" -> _P
  | "Q" -> _Q
  | "R" -> _R
  | "S" -> _S
  | "T" -> _T
  | "U" -> _U
  | "V" -> _V
  | "W" -> _W
  | "X" -> _X
  | "Y" -> _Y
  | "Z" -> _Z
  | "bracketleft" -> _bracketleft
  | "[" -> _bracketleft
  | "backslash" -> _backslash
  | "\\" -> _backslash
  | "bracketright" -> _bracketright
  | "]" -> _bracketright
  | "asciicircum" -> _asciicircum
  | "underscore" -> _underscore
  | "grave" -> _grave
  | "quoteleft" -> _quoteleft
  | "a" -> _a
  | "b" -> _b
  | "c" -> _c
  | "d" -> _d
  | "e" -> _e
  | "f" -> _f
  | "g" -> _g
  | "h" -> _h
  | "i" -> _i
  | "j" -> _j
  | "k" -> _k
  | "l" -> _l
  | "m" -> _m
  | "n" -> _n
  | "o" -> _o
  | "p" -> _p
  | "q" -> _q
  | "r" -> _r
  | "s" -> _s
  | "t" -> _t
  | "u" -> _u
  | "v" -> _v
  | "w" -> _w
  | "x" -> _x
  | "y" -> _y
  | "z" -> _z
  | "braceleft" -> _braceleft
  | "{" -> _braceleft
  | "bar" -> _bar
  | "|" -> _bar
  | "braceright" -> _braceright
  | "}" -> _braceright
  | "asciitilde" -> _asciitilde
  | "~" -> _asciitilde
  | ks -> raise (UnsupportedKeysym ks)

let string_of_keysym ks =
  let ksmap = [
    _VoidSymbol, "voidsymbol";
    _BackSpace, "backspace";
    _Tab, "tab";
    _ISO_Left_Tab, "left_tab";
    _Linefeed, "linefeed";
    _Clear, "clear";
    _Return, "return";
    _Pause, "pause";
    _Scroll_Lock, "scroll_lock";
    _Sys_Req, "sys_req";
    _Escape, "escape";
    _Delete, "delete";
    _Multi_key, "multi_key";
    _Home, "home";
    _Left, "left";
    _Up, "up";
    _Right, "right";
    _Down, "down";
    _Prior, "prior";
    _Page_Up, "page_up";
    _Next, "next";
    _Page_Down, "page_down";
    _End, "end";
    _Begin, "begin";
    _Select, "select";
    _Print, "print";
    _Execute, "execute";
    _Insert, "insert";
    _Undo, "undo";
    _Redo, "redo";
    _Menu, "menu";
    _Find, "find";
    _Cancel, "cancel";
    _Help, "help";
    _Break, "break";
    _Mode_switch, "mode_switch";
    _script_switch, "script_switch";
    _Num_Lock, "num_lock";
    _KP_Space, "kp_space";
    _KP_Tab, "kp_tab";
    _KP_Enter, "kp_enter";
    _KP_F1, "kp_f1";
    _KP_F2, "kp_f2";
    _KP_F3, "kp_f3";
    _KP_F4, "kp_f4";
    _KP_Home, "kp_home";
    _KP_Left, "kp_left";
    _KP_Up, "kp_up";
    _KP_Right, "kp_right";
    _KP_Down, "kp_down";
    _KP_Prior, "kp_prior";
    _KP_Page_Up, "kp_page_up";
    _KP_Next, "kp_next";
    _KP_Page_Down, "kp_page_down";
    _KP_End, "kp_end";
    _KP_Begin, "kp_begin";
    _KP_Insert, "kp_insert";
    _KP_Delete, "kp_delete";
    _KP_Equal, "kp_equal";
    _KP_Multiply, "kp_multiply";
    _KP_Add, "kp_add";
    _KP_Separator, "kp_separator";
    _KP_Subtract, "kp_subtract";
    _KP_Decimal, "kp_decimal";
    _KP_Divide, "kp_divide";
    _KP_0, "kp_0";
    _KP_1, "kp_1";
    _KP_2, "kp_2";
    _KP_3, "kp_3";
    _KP_4, "kp_4";
    _KP_5, "kp_5";
    _KP_6, "kp_6";
    _KP_7, "kp_7";
    _KP_8, "kp_8";
    _KP_9, "kp_9";
    _F1, "f1";
    _F2, "f2";
    _F3, "f3";
    _F4, "f4";
    _F5, "f5";
    _F6, "f6";
    _F7, "f7";
    _F8, "f8";
    _F9, "f9";
    _F10, "f10";
    _F11, "f11";
    _L1, "l1";
    _F12, "f12";
    _L2, "l2";
    _F13, "f13";
    _L3, "l3";
    _F14, "f14";
    _L4, "l4";
    _F15, "f15";
    _L5, "l5";
    _F16, "f16";
    _L6, "l6";
    _F17, "f17";
    _L7, "l7";
    _F18, "f18";
    _L8, "l8";
    _F19, "f19";
    _L9, "l9";
    _F20, "f20";
    _L10, "l10";
    _F21, "f21";
    _R1, "r1";
    _F22, "f22";
    _R2, "r2";
    _F23, "f23";
    _R3, "r3";
    _F24, "f24";
    _R4, "r4";
    _F25, "f25";
    _R5, "r5";
    _F26, "f26";
    _R6, "r6";
    _F27, "f27";
    _R7, "r7";
    _F28, "f28";
    _R8, "r8";
    _F29, "f29";
    _R9, "r9";
    _F30, "f30";
    _R10, "r10";
    _F31, "f31";
    _R11, "r11";
    _F32, "f32";
    _R12, "r12";
    _F33, "f33";
    _R13, "r13";
    _F34, "f34";
    _R14, "r14";
    _F35, "f35";
    _R15, "r15";
    _Shift_L, "shift_l";
    _Shift_R, "shift_r";
    _Control_L, "control_l";
    _Control_R, "control_r";
    _Caps_Lock, "caps_lock";
    _Shift_Lock, "shift_lock";
    _Meta_L, "meta_l";
    _Meta_R, "meta_r";
    _Alt_L, "alt_l";
    _Alt_R, "alt_r";
    _Super_L, "super_l";
    _Super_R, "super_r";
    _Hyper_L, "hyper_l";
    _Hyper_R, "hyper_r";
    _space, "space";
    _exclam, "exclam";
    _quotedbl, "quotedbl";
    _quotedbl, "\"";
    _numbersign, "numbersign";
    _numbersign, "#";
    _dollar, "dollar";
    _dollar, "$";
    _percent, "percent";
    _percent, "%";
    _ampersand, "ampersand";
    _ampersand, "&";
    _apostrophe, "apostrophe";
    _apostrophe, "'";
    _quoteright, "quoteright";
    _quoteright, "`";
    _parenleft, "parenleft";
    _parenleft, "(";
    _parenright, "parenright";
    _parenright, ")";
    _asterisk, "asterisk";
    _asterisk, "*";
    _plus, "plus";
    _plus, "+";
    _comma, "comma";
    _comma, ",";
    _minus, "minus";
    _minus, "-";
    _period, "period";
    _period, ".";
    _slash, "slash";
    _slash, "/";
    _0, "0";
    _1, "1";
    _2, "2";
    _3, "3";
    _4, "4";
    _5, "5";
    _6, "6";
    _7, "7";
    _8, "8";
    _9, "9";
    _colon, "colon";
    _semicolon, "semicolon";
    _semicolon, ";";
    _less, "less";
    _less, "<";
    _equal, "equal";
    _equal, "=";
    _greater, "greater";
    _greater, ">";
    _question, "question";
    _question, "?";
    _at, "at";
    _A, "A";
    _B, "B";
    _C, "C";
    _D, "D";
    _E, "E";
    _F, "F";
    _G, "G";
    _H, "H";
    _I, "I";
    _J, "J";
    _K, "K";
    _L, "L";
    _M, "M";
    _N, "N";
    _O, "O";
    _P, "P";
    _Q, "Q";
    _R, "R";
    _S, "S";
    _T, "T";
    _U, "U";
    _V, "V";
    _W, "W";
    _X, "X";
    _Y, "Y";
    _Z, "Z";
    _bracketleft, "bracketleft";
    _bracketleft, "[";
    _backslash, "backslash";
    _backslash, "\\";
    _bracketright, "bracketright";
    _bracketright, "]";
    _asciicircum, "asciicircum";
    _underscore, "underscore";
    _grave, "grave";
    _quoteleft, "quoteleft";
    _a, "a";
    _b, "b";
    _c, "c";
    _d, "d";
    _e, "e";
    _f, "f";
    _g, "g";
    _h, "h";
    _i, "i";
    _j, "j";
    _k, "k";
    _l, "l";
    _m, "m";
    _n, "n";
    _o, "o";
    _p, "p";
    _q, "q";
    _r, "r";
    _s, "s";
    _t, "t";
    _u, "u";
    _v, "v";
    _w, "w";
    _x, "x";
    _y, "y";
    _z, "z";
    _braceleft, "braceleft";
    _braceleft, "{";
    _bar, "bar";
    _bar, "|";
    _braceright, "braceright";
    _braceright, "}";
    _asciitilde, "asciitilde";
    _asciitilde, "~"
  ] in
  try List.assoc ks ksmap
  with Not_found -> raise (UnsupportedKeysym (string_of_int ks))

external usleep : int -> unit = "ml_usleep"

let sleep secs =
  usleep (int_of_float (secs *. 1000000.))
(*
let sleep secs =
  ignore (Unix.select [] [] [] secs)
*)
